from __future__ import annotations

import random
from typing import Any, Iterable, Sequence
from warnings import warn

from pydantic import AfterValidator, Field, model_validator
from typing_extensions import Annotated, Self

from albumentations.core.transforms_interface import BaseTransformInitSchema, DualTransform
from albumentations.core.types import MIN_UNIT_SIZE, PAIR, ColorType, Targets

from . import functional as fdropout


import numpy as np

from albumentations.core.pydantic import check_0plus, check_1plus, nondecreasing

__all__ = ["GridDropout"]


class GridDropout(DualTransform):
    """GridDropout, drops out rectangular regions of an image and the corresponding mask in a grid fashion.

    Args:
        ratio (float): The ratio of the mask holes to the unit_size (same for horizontal and vertical directions).
            Must be between 0 and 1. Default: 0.5.
        random_offset (bool): Whether to offset the grid randomly between 0 and grid unit size - hole size.
            If True, entered shift_x and shift_y are ignored and set randomly. Default: False.
        fill_value (Optional[ColorType]): Value for the dropped pixels. Default: 0.
        mask_fill_value (Optional[ColorType]): Value for the dropped pixels in mask.
            If None, transformation is not applied to the mask. Default: None.
        unit_size_range (Optional[tuple[int, int]]): Range from which to sample grid size. Default: None.
             Must be between 2 and the image shorter edge.
        holes_number_xy (Optional[tuple[int, int]]): The number of grid units in x and y directions.
            First value should be between 1 and image width//2,
            Second value should be between 1 and image height//2.
            Default: None.
        shift_xy (tuple[int, int]): Offsets of the grid start in x and y directions.
            Offsets of the grid start in x and y directions from (0,0) coordinate.
            Default: (0, 0).

        p (float): Probability of applying the transform. Default: 0.5.

    Targets:
        image, mask

    Image types:
        uint8, float32

    Reference:
        https://arxiv.org/abs/2001.04086

    """

    _targets = (Targets.IMAGE, Targets.MASK)

    class InitSchema(BaseTransformInitSchema):
        ratio: float = Field(description="The ratio of the mask holes to the unit_size.", gt=0, le=1)

        unit_size_min: int | None = Field(None, description="Minimum size of the grid unit.", ge=2)
        unit_size_max: int | None = Field(None, description="Maximum size of the grid unit.", ge=2)

        holes_number_x: int | None = Field(None, description="The number of grid units in x direction.", ge=1)
        holes_number_y: int | None = Field(None, description="The number of grid units in y direction.", ge=1)

        shift_x: int | None = Field(0, description="Offsets of the grid start in x direction.", ge=0)
        shift_y: int | None = Field(0, description="Offsets of the grid start in y direction.", ge=0)

        random_offset: bool = Field(False, description="Whether to offset the grid randomly.")
        fill_value: ColorType | None = Field(0, description="Value for the dropped pixels.")
        mask_fill_value: ColorType | None = Field(None, description="Value for the dropped pixels in mask.")
        unit_size_range: (
            Annotated[tuple[int, int], AfterValidator(check_1plus), AfterValidator(nondecreasing)] | None
        ) = None
        shift_xy: Annotated[tuple[int, int], AfterValidator(check_0plus)] = Field(
            (0, 0),
            description="Offsets of the grid start in x and y directions.",
        )
        holes_number_xy: Annotated[tuple[int, int], AfterValidator(check_1plus)] | None = Field(
            None,
            description="The number of grid units in x and y directions.",
        )

        @model_validator(mode="after")
        def validate_normalization(self) -> Self:
            if self.unit_size_min is not None and self.unit_size_max is not None:
                self.unit_size_range = self.unit_size_min, self.unit_size_max
                warn(
                    "unit_size_min and unit_size_max are deprecated. Use unit_size_range instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            if self.shift_x is not None and self.shift_y is not None:
                self.shift_xy = self.shift_x, self.shift_y
                warn("shift_x and shift_y are deprecated. Use shift_xy instead.", DeprecationWarning, stacklevel=2)

            if self.holes_number_x is not None and self.holes_number_y is not None:
                self.holes_number_xy = self.holes_number_x, self.holes_number_y
                warn(
                    "holes_number_x and holes_number_y are deprecated. Use holes_number_xy instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            if self.unit_size_range and not MIN_UNIT_SIZE <= self.unit_size_range[0] <= self.unit_size_range[1]:
                raise ValueError("Max unit size should be >= min size, both at least 2 pixels.")

            return self

    def __init__(
        self,
        ratio: float = 0.5,
        unit_size_min: int | None = None,
        unit_size_max: int | None = None,
        holes_number_x: int | None = None,
        holes_number_y: int | None = None,
        shift_x: int | None = None,
        shift_y: int | None = None,
        random_offset: bool = False,
        fill_value: ColorType = 0,
        mask_fill_value: ColorType | None = None,
        unit_size_range: tuple[int, int] | None = None,
        holes_number_xy: tuple[int, int] | None = None,
        shift_xy: tuple[int, int] = (0, 0),
        always_apply: bool | None = None,
        p: float = 0.5,
    ):
        super().__init__(p, always_apply)
        self.ratio = ratio
        self.unit_size_range = unit_size_range
        self.holes_number_xy = holes_number_xy
        self.random_offset = random_offset
        self.fill_value = fill_value
        self.mask_fill_value = mask_fill_value
        self.shift_xy = shift_xy

    def apply(self, img: np.ndarray, holes: Iterable[tuple[int, int, int, int]], **params: Any) -> np.ndarray:
        return fdropout.cutout(img, holes, self.fill_value)

    def apply_to_mask(
        self,
        mask: np.ndarray,
        holes: Iterable[tuple[int, int, int, int]],
        **params: Any,
    ) -> np.ndarray:
        if self.mask_fill_value is None:
            return mask

        return fdropout.cutout(mask, holes, self.mask_fill_value)

    def get_params_dependent_on_data(self, params: dict[str, Any], data: dict[str, Any]) -> dict[str, Any]:
        height, width = params["shape"][:2]
        unit_width, unit_height = self._calculate_unit_dimensions(width, height)
        hole_width, hole_height = self._calculate_hole_dimensions(unit_width, unit_height)
        shift_x, shift_y = self._calculate_shifts(unit_width, unit_height, hole_width, hole_height)
        holes = self._generate_holes(width, height, unit_width, unit_height, hole_width, hole_height, shift_x, shift_y)
        return {"holes": holes}

    def _calculate_unit_dimensions(self, width: int, height: int) -> tuple[int, int]:
        """Calculates the dimensions of the grid units."""
        if self.unit_size_range is not None:
            self._validate_unit_sizes(height, width)
            unit_size = random.randint(*self.unit_size_range)
            return unit_size, unit_size

        return self._calculate_dimensions_based_on_holes(width, height)

    def _validate_unit_sizes(self, height: int, width: int) -> None:
        """Validates the minimum and maximum unit sizes."""
        if self.unit_size_range is None:
            raise ValueError("unit_size_range must not be None.")
        if self.unit_size_range[1] > min(height, width):
            msg = "Grid size limits must be within the shortest image edge."
            raise ValueError(msg)

    def _calculate_dimensions_based_on_holes(self, width: int, height: int) -> tuple[int, int]:
        """Calculates dimensions based on the number of holes specified."""
        holes_number_x, holes_number_y = self.holes_number_xy or (None, None)
        unit_width = self._calculate_dimension(width, holes_number_x, 10)
        unit_height = self._calculate_dimension(height, holes_number_y, unit_width)
        return unit_width, unit_height

    @staticmethod
    def _calculate_dimension(dimension: int, holes_number: int | None, fallback: int) -> int:
        """Helper function to calculate unit width or height."""
        if holes_number is None:
            return max(2, dimension // fallback)

        if not 1 <= holes_number <= dimension // 2:
            raise ValueError(f"The number of holes must be between 1 and {dimension // 2}.")
        return dimension // holes_number

    def _calculate_hole_dimensions(self, unit_width: int, unit_height: int) -> tuple[int, int]:
        """Calculates the dimensions of the holes to be dropped out."""
        hole_width = int(unit_width * self.ratio)
        hole_height = int(unit_height * self.ratio)
        hole_width = min(max(hole_width, 1), unit_width - 1)
        hole_height = min(max(hole_height, 1), unit_height - 1)
        return hole_width, hole_height

    def _calculate_shifts(
        self,
        unit_width: int,
        unit_height: int,
        hole_width: int,
        hole_height: int,
    ) -> tuple[int, int]:
        """Calculates the shifts for the grid start."""
        if self.random_offset:
            shift_x = random.randint(0, unit_width - hole_width)
            shift_y = random.randint(0, unit_height - hole_height)
            return shift_x, shift_y

        if isinstance(self.shift_xy, Sequence) and len(self.shift_xy) == PAIR:
            shift_x = min(max(0, self.shift_xy[0]), unit_width - hole_width)
            shift_y = min(max(0, self.shift_xy[1]), unit_height - hole_height)
            return shift_x, shift_y

        return 0, 0

    def _generate_holes(
        self,
        width: int,
        height: int,
        unit_width: int,
        unit_height: int,
        hole_width: int,
        hole_height: int,
        shift_x: int,
        shift_y: int,
    ) -> list[tuple[int, int, int, int]]:
        """Generates the list of holes to be dropped out."""
        holes = []
        for i in range(width // unit_width + 1):
            for j in range(height // unit_height + 1):
                x1 = min(shift_x + unit_width * i, width)
                y1 = min(shift_y + unit_height * j, height)
                x2 = min(x1 + hole_width, width)
                y2 = min(y1 + hole_height, height)
                holes.append((x1, y1, x2, y2))
        return holes

    def get_transform_init_args_names(self) -> tuple[str, ...]:
        return (
            "ratio",
            "unit_size_range",
            "holes_number_xy",
            "shift_xy",
            "random_offset",
            "fill_value",
            "mask_fill_value",
        )



import logging
import pytest
from unittest.mock import MagicMock
import urllib.error

# Correct import statements based on your project structure
from albumentations.check_version import fetch_version_info, parse_version, check_for_updates, SUCCESS_HTML_CODE

@pytest.fixture
def mock_response_success():
    mock_response = MagicMock()
    mock_response.__enter__.return_value.read.return_value = b'{"info": {"version": "1.0.1"}}'
    mock_response.__enter__.return_value.info.return_value.get_content_charset.return_value = 'utf-8'
    mock_response.__enter__.return_value.status = SUCCESS_HTML_CODE
    return mock_response

@pytest.fixture
def mock_response_failure():
    mock_response = MagicMock()
    mock_response.__enter__.return_value.read.return_value = b'{}'
    mock_response.__enter__.return_value.info.return_value.get_content_charset.return_value = 'utf-8'
    mock_response.__enter__.return_value.status = 500  # use a different status code for failure
    return mock_response

def test_fetch_version_info_success(mocker, mock_response_success, caplog):
    mocker.patch('urllib.request.OpenerDirector.open', return_value=mock_response_success)
    with caplog.at_level(logging.DEBUG):
        result = fetch_version_info()
        assert "1.0.1" in result, "Should return version data when HTTP status is 200"

def test_fetch_version_info_failure(mocker, mock_response_failure):
    mocker.patch('urllib.request.OpenerDirector.open', return_value=mock_response_failure)
    result = fetch_version_info()
    assert result == "", "Should return empty string on HTTP failure"

def test_fetch_version_info_timeout(mocker, caplog):
    mocker.patch('urllib.request.OpenerDirector.open', side_effect=urllib.error.URLError("timeout"))
    result = fetch_version_info()
    assert result == "", "Should return empty string on timeout error"

def test_check_for_updates_new_version_available(mocker):
    mocker.patch('albumentations.check_version.fetch_version_info', return_value='{"info": {"version": "1.0.2"}}')
    mocker.patch('albumentations.check_version.current_version', '1.0.1')
    log_mock = mocker.patch('logging.Logger.info')
    check_for_updates()
    log_mock.assert_called_once_with(
        'A new version of Albumentations is available: 1.0.2 (you have 1.0.1). Upgrade using: pip install -U albumentations. To disable automatic update checks, set the environment variable NO_ALBUMENTATIONS_UPDATE to 1.'
    )

def test_check_for_updates_no_update_needed(mocker):
    mocker.patch('albumentations.check_version.fetch_version_info', return_value='{"info": {"version": "1.0.1"}}')
    mocker.patch('albumentations.check_version.current_version', '1.0.1')
    log_mock = mocker.patch('logging.Logger.info')
    check_for_updates()
    log_mock.assert_not_called()

def test_check_for_updates_exception_handling(mocker):
    mocker.patch('albumentations.check_version.fetch_version_info', side_effect=Exception("Error"))
    log_mock = mocker.patch('logging.Logger.info')
    check_for_updates()
    log_mock.assert_called_once_with("Failed to check for updates due to an unexpected error: Error. To disable automatic update checks, set the environment variable NO_ALBUMENTATIONS_UPDATE to 1.")

def test_parse_version_correct_input():
    json_input = '{"info": {"version": "1.0.2"}}'
    expected_version = "1.0.2"
    assert parse_version(json_input) == expected_version, "Should correctly parse the version from valid JSON data"

def test_parse_version_empty_input():
    json_input = ''
    assert parse_version(json_input) == "", "Should return an empty string when provided with an empty input"

def test_parse_version_malformed_json():
    json_input = '{"info": {}}'  # Malformed JSON
    assert parse_version(json_input) == "", "Should handle malformed JSON data gracefully by returning an empty string"

def test_parse_version_no_version_field():
    json_input = '{"info": {"release": "2021"}}'  # Missing 'version' field
    assert parse_version(json_input) == "", "Should return an empty string when 'version' field is missing"



from __future__ import annotations

from typing import Sequence, cast

import cv2
from albucore.utils import maybe_process_in_chunks, preserve_channel_dim

from albumentations.augmentations.geometric import functional as fgeometric
from albumentations.core.bbox_utils import denormalize_bbox, normalize_bbox
from albumentations.core.types import BoxInternalType, ColorType, KeypointInternalType


import numpy as np

__all__ = [
    "get_crop_coords",
    "crop_bbox_by_coords",
    "crop_keypoint_by_coords",
    "crop",
    "crop_and_pad",
    "crop_and_pad_bbox",
    "crop_and_pad_keypoint",
]


def get_crop_coords(
    height: int,
    width: int,
    crop_height: int,
    crop_width: int,
    h_start: float,
    w_start: float,
) -> tuple[int, int, int, int]:
    # h_start is [0, 1) and should map to [0, (height - crop_height)]  (note inclusive)
    # This is conceptually equivalent to mapping onto `range(0, (height - crop_height + 1))`
    # See: https://github.com/albumentations-team/albumentations/pull/1080
    # We want range for coordinated to be [0, image_size], right side is included
    y_min = int((height - crop_height + 1) * h_start)
    y_max = y_min + crop_height
    x_min = int((width - crop_width + 1) * w_start)
    x_max = x_min + crop_width
    return x_min, y_min, x_max, y_max


def crop_bbox_by_coords(
    bbox: BoxInternalType,
    crop_coords: tuple[int, int, int, int],
    rows: int,
    cols: int,
) -> BoxInternalType:
    denormalized_bbox = denormalize_bbox(bbox, rows, cols)

    x_min, y_min, x_max, y_max = denormalized_bbox[:4]
    x1, y1 = crop_coords[:2]
    cropped_bbox = x_min - x1, y_min - y1, x_max - x1, y_max - y1
    crop_height = crop_coords[3] - crop_coords[1]
    crop_width = crop_coords[2] - crop_coords[0]

    return cast(BoxInternalType, normalize_bbox(cropped_bbox, crop_height, crop_width))


def crop_keypoint_by_coords(
    keypoint: KeypointInternalType,
    crop_coords: tuple[int, int, int, int],
) -> KeypointInternalType:
    """Crop a keypoint using the provided coordinates of bottom-left and top-right corners in pixels and the
    required height and width of the crop.

    Args:
        keypoint (tuple): A keypoint `(x, y, angle, scale)`.
        crop_coords (tuple): Crop box coords `(x1, x2, y1, y2)`.

    Returns:
        A keypoint `(x, y, angle, scale)`.

    """
    x, y, angle, scale = keypoint[:4]
    x1, y1 = crop_coords[:2]
    return x - x1, y - y1, angle, scale


def get_center_crop_coords(height: int, width: int, crop_height: int, crop_width: int) -> tuple[int, int, int, int]:
    y_min = (height - crop_height) // 2
    y_max = y_min + crop_height
    x_min = (width - crop_width) // 2
    x_max = x_min + crop_width
    return x_min, y_min, x_max, y_max


def crop(img: np.ndarray, x_min: int, y_min: int, x_max: int, y_max: int) -> np.ndarray:
    height, width = img.shape[:2]
    if x_max <= x_min or y_max <= y_min:
        raise ValueError(
            "We should have x_min < x_max and y_min < y_max. But we got"
            f" (x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max})",
        )

    if x_min < 0 or x_max > width or y_min < 0 or y_max > height:
        raise ValueError(
            "Values for crop should be non negative and equal or smaller than image sizes"
            f"(x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max}, "
            f"height = {height}, width = {width})",
        )

    return img[y_min:y_max, x_min:x_max]


@preserve_channel_dim
def crop_and_pad(
    img: np.ndarray,
    crop_params: Sequence[int] | None,
    pad_params: Sequence[int] | None,
    pad_value: ColorType | None,
    rows: int,
    cols: int,
    interpolation: int,
    pad_mode: int,
    keep_size: bool,
) -> np.ndarray:
    if crop_params is not None and any(i != 0 for i in crop_params):
        img = crop(img, *crop_params)
    if pad_params is not None and any(i != 0 for i in pad_params):
        img = fgeometric.pad_with_params(
            img,
            pad_params[0],
            pad_params[1],
            pad_params[2],
            pad_params[3],
            border_mode=pad_mode,
            value=pad_value,
        )

    if keep_size:
        resize_fn = maybe_process_in_chunks(cv2.resize, dsize=(cols, rows), interpolation=interpolation)
        return resize_fn(img)

    return img


def crop_and_pad_bbox(
    bbox: BoxInternalType,
    crop_params: Sequence[int] | None,
    pad_params: Sequence[int] | None,
    rows: int,
    cols: int,
    result_rows: int,
    result_cols: int,
) -> BoxInternalType:
    x1, y1, x2, y2 = denormalize_bbox(bbox, rows, cols)[:4]

    if crop_params is not None:
        crop_x, crop_y = crop_params[:2]

        x1 -= crop_x
        y1 -= crop_y
        x2 -= crop_x
        y2 -= crop_y

    if pad_params is not None:
        top = pad_params[0]
        left = pad_params[2]

        x1 += left
        y1 += top
        x2 += left
        y2 += top

    return cast(BoxInternalType, normalize_bbox((x1, y1, x2, y2), result_rows, result_cols))


def crop_and_pad_keypoint(
    keypoint: KeypointInternalType,
    crop_params: Sequence[int] | None,
    pad_params: Sequence[int] | None,
    rows: int,
    cols: int,
    result_rows: int,
    result_cols: int,
    keep_size: bool,
) -> KeypointInternalType:
    x, y, angle, scale = keypoint[:4]

    if crop_params is not None:
        crop_x1, crop_y1 = crop_params[:2]
        x, y = x - crop_x1, y - crop_y1
    if pad_params is not None:
        top = pad_params[0]
        left = pad_params[2]

        x += left
        y += top

    if keep_size and (result_cols != cols or result_rows != rows):
        scale_x = cols / result_cols
        scale_y = rows / result_rows
        return fgeometric.keypoint_scale((x, y, angle, scale), scale_x, scale_y)

    return x, y, angle, scale



from __future__ import annotations

from typing import Any

import numpy as np
import torch

from albumentations.core.transforms_interface import BasicTransform
from albumentations.core.types import MONO_CHANNEL_DIMENSIONS, NUM_MULTI_CHANNEL_DIMENSIONS, Targets

__all__ = ["ToTensorV2"]


class ToTensorV2(BasicTransform):
    """Converts images/masks to PyTorch Tensors, inheriting from BasicTransform. Supports images in numpy `HWC` format
    and converts them to PyTorch `CHW` format. If the image is in `HW` format, it will be converted to PyTorch `HW`.

    Attributes:
        transpose_mask (bool): If True, transposes 3D input mask dimensions from `[height, width, num_channels]` to
            `[num_channels, height, width]`.
        always_apply (bool): Deprecated. Default: None.
        p (float): Probability of applying the transform. Default: 1.0.

    """

    _targets = (Targets.IMAGE, Targets.MASK)

    def __init__(self, transpose_mask: bool = False, p: float = 1.0, always_apply: bool | None = None):
        super().__init__(p=p, always_apply=always_apply)
        self.transpose_mask = transpose_mask

    @property
    def targets(self) -> dict[str, Any]:
        return {"image": self.apply, "mask": self.apply_to_mask, "masks": self.apply_to_masks}

    def apply(self, img: np.ndarray, **params: Any) -> torch.Tensor:
        if len(img.shape) not in [2, 3]:
            msg = "Albumentations only supports images in HW or HWC format"
            raise ValueError(msg)

        if len(img.shape) == MONO_CHANNEL_DIMENSIONS:
            img = np.expand_dims(img, 2)

        return torch.from_numpy(img.transpose(2, 0, 1))

    def apply_to_mask(self, mask: np.ndarray, **params: Any) -> torch.Tensor:
        if self.transpose_mask and mask.ndim == NUM_MULTI_CHANNEL_DIMENSIONS:
            mask = mask.transpose(2, 0, 1)
        return torch.from_numpy(mask)

    def apply_to_masks(self, masks: list[np.ndarray], **params: Any) -> list[torch.Tensor]:
        return [self.apply_to_mask(mask, **params) for mask in masks]

    def get_transform_init_args_names(self) -> tuple[str, ...]:
        return ("transpose_mask",)
